@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2015-2020 Canonical Ltd
+ * Copyright (C) 2015-2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -27,8 +27,10 @@ import (
 
 	"github.com/gorilla/mux"
 
+	"github.com/snapcore/snapd/overlord/aspectstate"
 	"github.com/snapcore/snapd/overlord/assertstate"
 	"github.com/snapcore/snapd/overlord/auth"
+	"github.com/snapcore/snapd/overlord/configstate"
 	"github.com/snapcore/snapd/overlord/snapstate"
 	"github.com/snapcore/snapd/overlord/state"
 	"github.com/snapcore/snapd/strutil"
@@ -57,6 +59,7 @@ var api = []*Command{
 	snapctlCmd,
 	usersCmd,
 	sectionsCmd,
+	categoriesCmd,
 	aliasesCmd,
 	appsCmd,
 	logsCmd,
@@ -79,6 +82,7 @@ var api = []*Command{
 	systemRecoveryKeysCmd,
 	quotaGroupsCmd,
 	quotaGroupInfoCmd,
+	aspectsCmd,
 }
 
 const (
@@ -131,21 +135,31 @@ func storeFrom(d *Daemon) snapstate.StoreService {
 }
 
 var (
-	snapstateInstall           = snapstate.Install
-	snapstateInstallPath       = snapstate.InstallPath
-	snapstateInstallPathMany   = snapstate.InstallPathMany
-	snapstateRefreshCandidates = snapstate.RefreshCandidates
-	snapstateTryPath           = snapstate.TryPath
-	snapstateUpdate            = snapstate.Update
-	snapstateUpdateMany        = snapstate.UpdateMany
-	snapstateInstallMany       = snapstate.InstallMany
-	snapstateRemoveMany        = snapstate.RemoveMany
-	snapstateRevert            = snapstate.Revert
-	snapstateRevertToRevision  = snapstate.RevertToRevision
-	snapstateSwitch            = snapstate.Switch
+	snapstateInstall                        = snapstate.Install
+	snapstateInstallPath                    = snapstate.InstallPath
+	snapstateInstallPathMany                = snapstate.InstallPathMany
+	snapstateRefreshCandidates              = snapstate.RefreshCandidates
+	snapstateTryPath                        = snapstate.TryPath
+	snapstateUpdate                         = snapstate.Update
+	snapstateUpdateMany                     = snapstate.UpdateMany
+	snapstateInstallMany                    = snapstate.InstallMany
+	snapstateRemoveMany                     = snapstate.RemoveMany
+	snapstateResolveValSetsEnforcementError = snapstate.ResolveValidationSetsEnforcementError
+	snapstateRevert                         = snapstate.Revert
+	snapstateRevertToRevision               = snapstate.RevertToRevision
+	snapstateSwitch                         = snapstate.Switch
+	snapstateProceedWithRefresh             = snapstate.ProceedWithRefresh
+	snapstateHoldRefreshesBySystem          = snapstate.HoldRefreshesBySystem
+	snapstateLongestGatingHold              = snapstate.LongestGatingHold
+	snapstateSystemHold                     = snapstate.SystemHold
+
+	configstateConfigureInstalled = configstate.ConfigureInstalled
 
 	assertstateRefreshSnapAssertions         = assertstate.RefreshSnapAssertions
 	assertstateRestoreValidationSetsTracking = assertstate.RestoreValidationSetsTracking
+
+	aspectstateGet = aspectstate.Get
+	aspectstateSet = aspectstate.Set
 )
 
 func ensureStateSoonImpl(st *state.State) {
