@@ -65,7 +65,7 @@ var (
 	kernelFeatures        = apparmor_sandbox.KernelFeatures
 	parserFeatures        = apparmor_sandbox.ParserFeatures
 	loadProfiles          = apparmor_sandbox.LoadProfiles
-	unloadProfiles        = apparmor_sandbox.UnloadProfiles
+	removeCachedProfiles  = apparmor_sandbox.RemoveCachedProfiles
 
 	// make sure that apparmor profile fulfills the late discarding backend
 	// interface
@@ -101,15 +101,6 @@ func (b *Backend) Initialize(opts *interfaces.SecurityBackendOptions) error {
 	// possible because snapd must be able to install a new core and only at
 	// that moment generate it.
 
-	// Inspect the system and sets up local apparmor policy for snap-confine.
-	// Local policy is included by the system-wide policy. If the local policy
-	// has changed then the apparmor profile for snap-confine is reloaded.
-
-	// Create the local policy directory if it is not there.
-	if err := os.MkdirAll(dirs.SnapConfineAppArmorDir, 0755); err != nil {
-		return fmt.Errorf("cannot create snap-confine policy directory: %s", err)
-	}
-
 	// Check the /proc/self/exe symlink, this is needed below but we want to
 	// fail early if this fails for whatever reason.
 	exe, err := os.Readlink(procSelfExe)
@@ -117,56 +108,8 @@ func (b *Backend) Initialize(opts *interfaces.SecurityBackendOptions) error {
 		return fmt.Errorf("cannot read %s: %s", procSelfExe, err)
 	}
 
-	// Location of the generated policy.
-	glob := "*"
-	policy := make(map[string]osutil.FileState)
-
-	// Check if NFS is mounted at or under $HOME. Because NFS is not
-	// transparent to apparmor we must alter our profile to counter that and
-	// allow snap-confine to work.
-	if nfs, err := osutil.IsHomeUsingNFS(); err != nil {
-		logger.Noticef("cannot determine if NFS is in use: %v", err)
-	} else if nfs {
-		policy["nfs-support"] = &osutil.MemoryFileState{
-			Content: []byte(nfsSnippet),
-			Mode:    0644,
-		}
-		logger.Noticef("snapd enabled NFS support, additional implicit network permissions granted")
-	}
-
-	// Check if '/' is on overlayfs. If so, add the necessary rules for
-	// upperdir and allow snap-confine to work.
-	if overlayRoot, err := isRootWritableOverlay(); err != nil {
-		logger.Noticef("cannot determine if root filesystem on overlay: %v", err)
-	} else if overlayRoot != "" {
-		snippet := strings.Replace(overlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
-		policy["overlay-root"] = &osutil.MemoryFileState{
-			Content: []byte(snippet),
-			Mode:    0644,
-		}
-		logger.Noticef("snapd enabled root filesystem on overlay support, additional upperdir permissions granted")
-	}
-
-	// Check whether apparmor_parser supports bpf capability. Some older
-	// versions do not, hence the capability cannot be part of the default
-	// profile of snap-confine as loading it would fail.
-	if features, err := apparmor_sandbox.ParserFeatures(); err != nil {
-		logger.Noticef("cannot determine apparmor_parser features: %v", err)
-	} else if strutil.ListContains(features, "cap-bpf") {
-		policy["cap-bpf"] = &osutil.MemoryFileState{
-			Content: []byte(capabilityBPFSnippet),
-			Mode:    0644,
-		}
-	}
-
-	// Ensure that generated policy is what we computed above.
-	created, removed, err := osutil.EnsureDirState(dirs.SnapConfineAppArmorDir, glob, policy)
-	if err != nil {
-		return fmt.Errorf("cannot synchronize snap-confine policy: %s", err)
-	}
-	if len(created) == 0 && len(removed) == 0 {
-		// If the generated policy didn't change, we're all done.
-		return nil
+	if _, err := apparmor_sandbox.SetupSnapConfineSnippets(); err != nil {
+		return err
 	}
 
 	// If snapd is executing from the core snap the it means it has
@@ -202,7 +145,7 @@ func (b *Backend) Initialize(opts *interfaces.SecurityBackendOptions) error {
 		// When we cannot reload the profile then let's remove the generated
 		// policy. Maybe we have caused the problem so it's better to let other
 		// things work.
-		osutil.EnsureDirState(dirs.SnapConfineAppArmorDir, glob, nil)
+		apparmor_sandbox.RemoveSnapConfineSnippets()
 		return fmt.Errorf("cannot reload snap-confine apparmor profile: %v", err)
 	}
 	return nil
@@ -233,6 +176,10 @@ func snapConfineFromSnapProfile(info *snap.Info) (dir, glob string, content map[
 	patchedProfileText := bytes.Replace(
 		vanillaProfileText, []byte("/usr/lib/snapd/snap-confine"), []byte(snapConfineInCore), -1)
 
+	// Replace the path to the vanilla snap-confine apparmor snippets
+	patchedProfileText = bytes.Replace(
+		patchedProfileText, []byte("/var/lib/snapd/apparmor/snap-confine"), []byte(apparmor_sandbox.SnapConfineAppArmorDir), -1)
+
 	// Also replace the test providing access to verbatim
 	// /usr/lib/snapd/snap-confine, which is necessary because to execute snaps
 	// from strict snaps, we need to be able read and map
@@ -289,7 +236,7 @@ func snapConfineProfileName(snapName string, rev snap.Revision) string {
 //
 // Additionally it will cleanup stale apparmor profiles it created.
 func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
-	if err := os.MkdirAll(dirs.SnapConfineAppArmorDir, 0755); err != nil {
+	if err := os.MkdirAll(apparmor_sandbox.SnapConfineAppArmorDir, 0755); err != nil {
 		return fmt.Errorf("cannot create snap-confine policy directory: %s", err)
 	}
 	dir, glob, content, err := snapConfineFromSnapProfile(info)
@@ -324,15 +271,15 @@ func (b *Backend) setupSnapConfineReexec(info *snap.Info) error {
 		aaFlags = apparmor_sandbox.SkipKernelLoad
 	}
 	errReload := loadProfiles(pathnames, cache, aaFlags)
-	errUnload := unloadProfiles(removed, cache)
+	errRemoveCached := removeCachedProfiles(removed, cache)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot synchronize snap-confine apparmor profile: %s", errEnsure)
 	}
 	if errReload != nil {
 		return fmt.Errorf("cannot reload snap-confine apparmor profile: %s", errReload)
 	}
-	if errUnload != nil {
-		return fmt.Errorf("cannot unload snap-confine apparmor profile: %s", errReload)
+	if errRemoveCached != nil {
+		return fmt.Errorf("cannot remove cached snap-confine apparmor profile: %s", errReload)
 	}
 	return nil
 }
@@ -441,7 +388,7 @@ func (b *Backend) prepareProfiles(snapInfo *snap.Info, opts interfaces.Confineme
 		return nil, fmt.Errorf("cannot create directory for apparmor profiles %q: %s", dir, err)
 	}
 	changed, removedPaths, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, content)
-	// XXX: in the old code this error was reported late, after doing load/unload.
+	// XXX: in the old code this error was reported late, after doing load/removeCached.
 	if errEnsure != nil {
 		return nil, fmt.Errorf("cannot synchronize security files for snap %q: %s", snapName, errEnsure)
 	}
@@ -507,14 +454,14 @@ func (b *Backend) Setup(snapInfo *snap.Info, opts interfaces.ConfinementOptions,
 	timings.Run(tm, "load-profiles[unchanged]", fmt.Sprintf("load unchanged security profiles of snap %q", snapInfo.InstanceName()), func(nesttm timings.Measurer) {
 		errReloadOther = loadProfiles(prof.unchanged, apparmor_sandbox.CacheDir, aaFlags)
 	})
-	errUnload := unloadProfiles(prof.removed, apparmor_sandbox.CacheDir)
+	errRemoveCached := removeCachedProfiles(prof.removed, apparmor_sandbox.CacheDir)
 	if errReloadChanged != nil {
 		return errReloadChanged
 	}
 	if errReloadOther != nil {
 		return errReloadOther
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 // SetupMany creates and loads apparmor profiles for multiple snaps.
@@ -558,7 +505,7 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			errReloadOther = loadProfiles(allUnchangedPaths, apparmor_sandbox.CacheDir, aaFlags)
 		})
 
-		errUnload := unloadProfiles(allRemovedPaths, apparmor_sandbox.CacheDir)
+		errRemoveCached := removeCachedProfiles(allRemovedPaths, apparmor_sandbox.CacheDir)
 		if errReloadChanged != nil {
 			logger.Noticef("failed to batch-reload changed profiles: %s", errReloadChanged)
 			fallback = true
@@ -567,8 +514,8 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 			logger.Noticef("failed to batch-reload unchanged profiles: %s", errReloadOther)
 			fallback = true
 		}
-		if errUnload != nil {
-			logger.Noticef("failed to batch-unload profiles: %s", errUnload)
+		if errRemoveCached != nil {
+			logger.Noticef("failed to batch-remove cached profiles: %s", errRemoveCached)
 			fallback = true
 		}
 	}
@@ -586,18 +533,40 @@ func (b *Backend) SetupMany(snaps []*snap.Info, confinement func(snapName string
 	return errors
 }
 
-// Remove removes and unloads apparmor profiles of a given snap.
+// Removes all AppArmor profiles from disk but does not unload them from the
+// kernel - currently it is not possible to ensure that all services of a snap
+// are stopped and so it is not possible to safely unload the profile for a snap
+// from the kernel and so we only remove it from the cache on disk
+func RemoveAllSnapAppArmorProfiles() error {
+	dir := dirs.SnapAppArmorDir
+	globs := []string{"snap.*", "snap-update-ns.*", "snap-confine.*"}
+	cache := apparmor_sandbox.CacheDir
+	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, nil)
+	errRemoveCached := removeCachedProfiles(removed, cache)
+	switch {
+	case errEnsure != nil && errRemoveCached != nil:
+		return fmt.Errorf("cannot remove apparmor profiles: %s (and also %s)", errEnsure, errRemoveCached)
+	case errEnsure != nil:
+		return fmt.Errorf("cannot remove apparmor profiles: %s", errEnsure)
+	case errRemoveCached != nil:
+		return errRemoveCached
+	default:
+		return nil
+	}
+}
+
+// Remove removes the apparmor profiles of a given snap from disk and the cache.
 func (b *Backend) Remove(snapName string) error {
 	dir := dirs.SnapAppArmorDir
 	globs := profileGlobs(snapName)
 	cache := apparmor_sandbox.CacheDir
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dir, globs, nil)
-	// always try to unload affected profiles
-	errUnload := unloadProfiles(removed, cache)
+	// always try to remove affected profiles from the cache
+	errRemoveCached := removeCachedProfiles(removed, cache)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot synchronize security files for snap %q: %s", snapName, errEnsure)
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type) error {
@@ -609,14 +578,15 @@ func (b *Backend) RemoveLate(snapName string, rev snap.Revision, typ snap.Type)
 
 	globs := []string{snapConfineProfileName(snapName, rev)}
 	_, removed, errEnsure := osutil.EnsureDirStateGlobs(dirs.SnapAppArmorDir, globs, nil)
-	// XXX: unloadProfiles() does not unload profiles from the kernel, but
-	// only removes profiles from the cache
-	// always try to unload the affected profile
-	errUnload := unloadProfiles(removed, apparmor_sandbox.CacheDir)
+	// XXX: we should also try and unload the profile from the kernel
+	// instead of just removing it from the cache but currently it is not
+	// possible to ensure all snap services are stopped at this time and so
+	// it is not safe to unload the profile
+	errRemoveCached := removeCachedProfiles(removed, apparmor_sandbox.CacheDir)
 	if errEnsure != nil {
 		return fmt.Errorf("cannot remove security profiles for snap %q (%s): %s", snapName, rev, errEnsure)
 	}
-	return errUnload
+	return errRemoveCached
 }
 
 var (
@@ -666,7 +636,7 @@ func addUpdateNSProfile(snapInfo *snap.Info, snippets string, content map[string
 			return snapInfo.InstanceName()
 		case "###SNIPPETS###":
 			if overlayRoot, _ := isRootWritableOverlay(); overlayRoot != "" {
-				snippets += strings.Replace(overlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
+				snippets += strings.Replace(apparmor_sandbox.OverlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
 			}
 			return snippets
 		}
@@ -853,16 +823,20 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
   %[2]s
 `, usrLibSnapdConfineTransitionRule, nonBaseCoreTransitionSnippet)
 
+		case "###INCLUDE_IF_EXISTS_SNAP_TUNING###":
+			features, _ := parserFeatures()
+			if strutil.ListContains(features, "include-if-exists") {
+				return `#include if exists "/var/lib/snapd/apparmor/snap-tuning"`
+			}
+			return ""
 		case "###VAR###":
 			return templateVariables(snapInfo, securityTag, cmdName)
 		case "###PROFILEATTACH###":
 			return fmt.Sprintf("profile \"%s\"", securityTag)
 		case "###CHANGEPROFILE_RULE###":
 			features, _ := parserFeatures()
-			for _, f := range features {
-				if f == "unsafe" {
-					return "change_profile unsafe /**,"
-				}
+			if strutil.ListContains(features, "unsafe") {
+				return "change_profile unsafe /**,"
 			}
 			return "change_profile,"
 		case "###SNIPPETS###":
@@ -882,11 +856,11 @@ func (b *Backend) addContent(securityTag string, snapInfo *snap.Info, cmdName st
 				// allow access to SNAP_USER_* files.
 				tagSnippets = snippetForTag
 				if nfs, _ := osutil.IsHomeUsingNFS(); nfs {
-					tagSnippets += nfsSnippet
+					tagSnippets += apparmor_sandbox.NfsSnippet
 				}
 
 				if overlayRoot, _ := isRootWritableOverlay(); overlayRoot != "" {
-					snippet := strings.Replace(overlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
+					snippet := strings.Replace(apparmor_sandbox.OverlayRootSnippet, "###UPPERDIR###", overlayRoot, -1)
 					tagSnippets += snippet
 				}
 
