@@ -1,7 +1,7 @@
 // -*- Mode: Go; indent-tabs-mode: t -*-
 
 /*
- * Copyright (C) 2021 Canonical Ltd
+ * Copyright (C) 2022 Canonical Ltd
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 3 as
@@ -518,6 +518,8 @@ func (s *ValidateSuite) TestAppStopMode(c *C) {
 		{"sigusr1-all", true},
 		{"sigusr2", true},
 		{"sigusr2-all", true},
+		{"sigint", true},
+		{"sigint-all", true},
 		// bad
 		{"invalid-thing", false},
 	} {
@@ -537,26 +539,32 @@ func (s *ValidateSuite) TestAppRefreshMode(c *C) {
 	// check services
 	for _, t := range []struct {
 		refreshMode string
-		ok          bool
+		daemon      string
+		errMsg      string
 	}{
 		// good
-		{"", true},
-		{"endure", true},
-		{"restart", true},
+		{"", "simple", ""},
+		{"endure", "simple", ""},
+		{"restart", "simple", ""},
+		{"ignore-running", "", ""},
 		// bad
-		{"invalid-thing", false},
+		{"invalid-thing", "simple", `"refresh-mode" field contains invalid value "invalid-thing"`},
+		{"endure", "", `"refresh-mode" for app "foo" can only have value "ignore-running"`},
+		{"restart", "", `"refresh-mode" for app "foo" can only have value "ignore-running"`},
+		{"ignore-running", "simple", `"refresh-mode" cannot be set to "ignore-running" for services`},
 	} {
-		err := ValidateApp(&AppInfo{Name: "foo", Daemon: "simple", DaemonScope: SystemDaemon, RefreshMode: t.refreshMode})
-		if t.ok {
+		var daemonScope DaemonScope
+		if t.daemon != "" {
+			daemonScope = SystemDaemon
+		}
+
+		err := ValidateApp(&AppInfo{Name: "foo", Daemon: t.daemon, DaemonScope: daemonScope, RefreshMode: t.refreshMode})
+		if t.errMsg == "" {
 			c.Check(err, IsNil)
 		} else {
-			c.Check(err, ErrorMatches, fmt.Sprintf(`"refresh-mode" field contains invalid value %q`, t.refreshMode))
+			c.Check(err, ErrorMatches, t.errMsg)
 		}
 	}
-
-	// non-services cannot have a refresh-mode
-	err := ValidateApp(&AppInfo{Name: "foo", Daemon: "", RefreshMode: "endure"})
-	c.Check(err, ErrorMatches, `"refresh-mode" cannot be used for "foo", only for services`)
 }
 
 func (s *ValidateSuite) TestAppWhitelistError(c *C) {
@@ -665,6 +673,28 @@ apps:
 
 // Validate
 
+func (s *ValidateSuite) TestDetectInvalidProvenance(c *C) {
+	info, err := InfoFromSnapYaml([]byte(`name: foo
+version: 1.0
+provenance: "--"
+`))
+	c.Assert(err, IsNil)
+
+	err = Validate(info)
+	c.Check(err, ErrorMatches, `invalid provenance: .*`)
+}
+
+func (s *ValidateSuite) TestDetectExplicitDefaultProvenance(c *C) {
+	info, err := InfoFromSnapYaml([]byte(`name: foo
+version: 1.0
+provenance: global-upload
+`))
+	c.Assert(err, IsNil)
+
+	err = Validate(info)
+	c.Check(err, ErrorMatches, `provenance cannot be set to default \(global-upload\) explicitly`)
+}
+
 func (s *ValidateSuite) TestDetectIllegalYamlBinaries(c *C) {
 	info, err := InfoFromSnapYaml([]byte(`name: foo
 version: 1.0
@@ -943,6 +973,16 @@ func (s *ValidateSuite) TestValidateLayout(c *C) {
 		ErrorMatches, `layout "/sys" in an off-limits area`)
 	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/run", Type: "tmpfs"}, nil),
 		ErrorMatches, `layout "/run" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/run/foo", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/run/foo" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/run/systemd", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/run/systemd" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/var/run", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/var/run" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/var/run/foo", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/var/run/foo" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/var/run/systemd", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/var/run/systemd" in an off-limits area`)
 	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/boot", Type: "tmpfs"}, nil),
 		ErrorMatches, `layout "/boot" in an off-limits area`)
 	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/lost+found", Type: "tmpfs"}, nil),
@@ -959,6 +999,10 @@ func (s *ValidateSuite) TestValidateLayout(c *C) {
 		ErrorMatches, `layout "/lib/firmware" in an off-limits area`)
 	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/lib/modules", Type: "tmpfs"}, nil),
 		ErrorMatches, `layout "/lib/modules" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/usr/lib/firmware", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/usr/lib/firmware" in an off-limits area`)
+	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/usr/lib/modules", Type: "tmpfs"}, nil),
+		ErrorMatches, `layout "/usr/lib/modules" in an off-limits area`)
 	c.Check(ValidateLayout(&Layout{Snap: si, Path: "/tmp", Type: "tmpfs"}, nil),
 		ErrorMatches, `layout "/tmp" in an off-limits area`)
 
